/* ERD - Copyright (C) 2005 bakunin - Andrea Marchesini 
 *                              <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "erd.h"

struct io_data
{
  int type;

  int port;
  char *server;
  char *unixpath;

  char *servername;
  int listen;

  int ssl;
  char *certificate;
  char *privatekey;

  char *command;
};

static struct io_data *io_parse (char *str);
static char *io_parse_str (char *str, int *ch, int len);
static int io_parse_int (char *str, int *ch, int len);

struct erd_config *
io_config (char *input, char *output)
{
  struct io_data *data;
  struct erd_config *conf;

  if (!(conf = (struct erd_config *) malloc (sizeof (struct erd_config))))
    fatal ("Error: memory.");

  memset (conf, 0, sizeof (struct erd_config));

  /* Input */
  data = io_parse (input);

  /* Unix */
  conf->input_type = data->type;
  if (data->type == ERD_UNIXSOCKET)
    {
      if (!data->unixpath)
	conf->input_unixpath = strdup (ERD_UNIX_SOCKET);
      else
	conf->input_unixpath = data->unixpath;
    }

  /* Tcp */
  else if (data->type == ERD_TCPSOCKET)
    {
      conf->input_port = data->port;

      if (!data->servername)
	conf->input_servername = data->servername;
    }

  else if (data->type == ERD_COMMAND)
    {
      if (!data->command)
	fatal ("Set the input command.");

      conf->input_command = data->command;
    }

  conf->input_listen = data->listen;
  conf->input_ssl = data->ssl;

  conf->input_certificate = data->certificate;
  conf->input_privatekey = data->privatekey;

  if (conf->input_ssl
      && (!conf->input_privatekey || !conf->input_certificate))
    fatal ("Set certificate and privatekey flags for ssl support.");

  /* Output */
  data = io_parse (output);

  /* Unix */
  conf->output_type = data->type;
  if (data->type == ERD_UNIXSOCKET)
    {
      if (!data->unixpath)
	conf->output_unixpath = strdup (ERD_UNIX_SOCKET);
      else
	conf->output_unixpath = data->unixpath;
    }

  /* Tcp */
  else if (data->type == ERD_TCPSOCKET)
    {
      conf->output_port = data->port;

      if (!data->server)
	conf->output_server = data->server;
      else
	conf->output_server = strdup ("localhost");
    }

  else if (data->type == ERD_COMMAND)
    {
      if (!data->command)
	fatal ("Set the output command.");

      conf->output_command = data->command;
    }

  conf->output_ssl = data->ssl;

  return conf;
}

static struct io_data *
io_parse (char *str)
{
  int ch = 0;
  int len = strlen (str);
  struct io_data *data;

  if (!(data = (struct io_data *) malloc (sizeof (struct io_data))))
    fatal ("Error: memory.");

  memset (data, 0, sizeof (struct io_data));

  while (ch < len)
    {

      switch (*(str + ch))
	{
	case 'c':
	  if (!strncmp (str + ch, "certificate=", 12))
	    {
	      ch += 12;
	      data->certificate = io_parse_str (str, &ch, len);
	    }

	  else if (!strncmp (str + ch, "command=", 8))
	    {
	      ch += 8;
	      data->command = io_parse_str (str, &ch, len);
	    }

	  else
	    fatal ("Error in parameter: %s", str + ch);
	  break;

	case 'l':
	  if (!strncmp (str + ch, "listen=", 7))
	    {
	      ch += 7;
	      data->listen = io_parse_int (str, &ch, len);
	    }

	  else
	    fatal ("Error in parameter: %s", str + ch);
	  break;

	case 'p':
	  if (!strncmp (str + ch, "port=", 5))
	    {
	      ch += 5;
	      data->port = io_parse_int (str, &ch, len);
	    }

	  else if (!strncmp (str + ch, "privatekey=", 11))
	    {
	      ch += 11;
	      data->privatekey = io_parse_str (str, &ch, len);
	    }

	  else
	    fatal ("Error in parameter: %s", str + ch);
	  break;

	case 's':
	  if (!strncmp (str + ch, "server=", 7))
	    {
	      ch += 7;
	      data->server = io_parse_str (str, &ch, len);
	    }

	  else if (!strncmp (str + ch, "servername=", 11))
	    {
	      ch += 11;
	      data->servername = io_parse_str (str, &ch, len);
	    }

	  else if (!strncmp (str + ch, "ssl=", 4))
	    {
	      ch += 4;
	      data->ssl = io_parse_int (str, &ch, len);
	    }

	  else
	    fatal ("Error in parameter: %s", str + ch);
	  break;

	case 't':
	  if (!strncmp (str + ch, "type=", 5))
	    {
	      char *c;
	      ch += 5;
	      c = io_parse_str (str, &ch, len);

	      data->type = config_type (c);
	    }

	  else
	    fatal ("Error in parameter: %s", str + ch);
	  break;

	case 'u':
	  if (!strncmp (str + ch, "unixpath=", 9))
	    {
	      ch += 9;
	      data->unixpath = io_parse_str (str, &ch, len);
	    }

	  else
	    fatal ("Error in parameter: %s", str + ch);
	  break;

	default:
	  fatal ("Error in parameter: %s", str + ch);
	  break;
	}


      ch++;
    }

  return data;
}


static char *
io_parse_str (char *str, int *ch, int len)
{
  char tmp[256];
  char *output;
  int i = 0;

  while (*(str + *ch) != ',' && *ch < len && i < 256)
    {
      *(tmp + i) = *(str + *ch);
      (*ch)++;
      i++;
    }

  *(tmp + i) = 0;

  if (!(output = strdup (tmp)))
    fatal ("Error: memory.");

  return output;
}

static int
io_parse_int (char *str, int *ch, int len)
{
  int output;

  output = atoi (str + *ch);

  while (*(str + *ch) != ',' && *(str + *ch) != ' ' && *ch < len)
    (*ch)++;

  return output;
}

/* EOF */
