/* ERD - Copyright (C) 2005 bakunin - Andrea Marchesini 
 *                              <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "erd.h"

static void usage (void);

void
fatal (char *text, ...)
{
  va_list va;
  char s[1024];

  va_start (va, text);

  vsnprintf (s, sizeof (s), text, va);

  if (log_is_opened ())
    log_write (ERD_ERROR, s);
  else
    fprintf (stderr, "%s\n", s);

  exit (1);
}

void
signal_off (void)
{
  signal (SIGPIPE, SIG_IGN);
}

void
signal_on (void)
{
  signal (SIGPIPE, SIG_DFL);
}

int
main (int argc, char **argv, char **env)
{
  struct erd_config *erd;
  char *config_file = NULL;
  char *logfile = NULL;
  char *input = NULL;
  char *output = NULL;
  int background = 0;
  int debug = 0;
  int fd, c_fd = 0;
  int i;

  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "-o") || !strcmp (argv[i], "--output"))
	{
	  if (!(output = argv[i + 1]))
	    {
	      usage ();
	      return 1;
	    }

	  i++;
	}

      else if (!strcmp (argv[i], "-i") || !strcmp (argv[i], "--input"))
	{
	  if (!(input = argv[i + 1]))
	    {
	      usage ();
	      return 1;
	    }

	  i++;
	}

      else if (!strcmp (argv[i], "-f") || !strcmp (argv[i], "--file"))
	{
	  if (!(config_file = argv[i + 1]))
	    {
	      usage ();
	      return 1;
	    }

	  i++;
	}

      else if (!strcmp (argv[i], "-l") || !strcmp (argv[i], "--logfile"))
	{
	  if (!(logfile = argv[i + 1]))
	    {
	      usage ();
	      return 1;
	    }

	  i++;
	}

      else if (!strcmp (argv[i], "-b") || !strcmp (argv[i], "--background"))
	{
	  if (!argv[i + 1])
	    {
	      usage ();
	      return 1;
	    }

	  background = atoi (argv[i + 1]);
	  i++;
	}

      else if (!strcmp (argv[i], "-d") || !strcmp (argv[i], "--debug"))
	{
	  if (!argv[i + 1])
	    {
	      usage ();
	      return 1;
	    }

	  debug = atoi (argv[i + 1]);
	  i++;
	}

      else if (!strcmp (argv[i], "-h") || !strcmp (argv[i], "--help"))
	{
	  usage ();
	  return 0;
	}

      else
	{
	  usage ();
	  return 1;
	}
    }

  if (!config_file)
    {
      if (!input && !output)
	config_file = ERD_CONFIG_FILE;
      else if (!input || !output)
	{
	  usage ();
	  return 1;
	}
    }

  if (config_file)
    erd = config (config_file);
  else
    erd = io_config (input, output);

  if (!erd)
    return 1;

  if (debug)
    erd->debug = debug;

  if (logfile)
    erd->logfile = logfile;

  if (background)
    erd->background = background;

  log_init (erd);

  if (erd->background)
    if (fork ())
      return 0;

  log_write (ERD_ERROR, "%s started", PACKAGE);

  signal_off ();

  if (erd->input_type == ERD_COMMAND || erd->input_type == ERD_FD)
    log_write (ERD_INFO, "Opening file descriptor...");
  else
    log_write (ERD_INFO, "Opening socket...");

  ssl_init ();

  if (erd->input_ssl)
    ssl_certificate (erd);

  signal (SIGPIPE, SIG_IGN);

  erd->env = env;

  if (erd->input_type == ERD_COMMAND)
    {
      command_action (erd);
      log_quit ();
      signal_on ();
      return 0;
    }

  else if (erd->input_type == ERD_FD)
    {
      struct erd_item input;
      struct termios t1, t2;

      input.fd = STDIN_FILENO;
      input.pid = 0;

      if ((input.use_ssl = erd->input_ssl))
	{
	  if (!(input.ssl = SSL_new (ssl_get_ctx ())))
	    fatal ("SSL create error.");

	  SSL_set_fd (input.ssl, input.fd);

	  if (SSL_accept (input.ssl) < 0)
	    {
	      log_write (ERD_ERROR, "SSL client error.");
	      log_quit ();
	      signal_on ();
	      return 1;
	    }
	}

      tcgetattr(STDIN_FILENO, &t1);
      t2=t1;

      t2.c_lflag &= ~(ISIG | ICANON | ECHO | ECHOE | ECHOK | ECHONL);

      tcsetattr (STDIN_FILENO, TCSADRAIN, &t2);

      erd_redirect (erd, &input);
      tcsetattr (STDIN_FILENO, TCSADRAIN, &t1);

      log_quit ();
      signal_on ();
      return 0;
    }

  else
    {
      fd = socket_open (erd);

      while (1)
	{
	  log_write (ERD_INFO, "Accepting connections...");

	  if (erd->input_type == ERD_UNIXSOCKET)
	    {
	      if ((c_fd = accept (fd, 0, 0)) < 0)
		fatal ("Accept error.");

	      log_write (ERD_WARN, "New client request (%d)", c_fd);
	    }

	  else if (erd->input_type == ERD_TCPSOCKET)
	    {
	      struct sockaddr_in incoming;
	      size_t size_in = sizeof (struct sockaddr_in);
	      int port, addr;

	      if ((c_fd =
		   accept (fd, (struct sockaddr *) &incoming, &size_in)) < 0)
		fatal ("Accept error.");

	      port = ntohs (incoming.sin_port);
	      addr = ntohl (incoming.sin_addr.s_addr);

	      log_write (ERD_WARN,
			 "New client request (%d - %d.%d.%d.%d:%5d)", c_fd,
			 (unsigned int) addr >> 24,
			 (unsigned int) (addr >> 16) % 256,
			 (unsigned int) (addr >> 8) % 256,
			 (unsigned int) addr % 256, port);

	    }

	  socket_new_client (erd, c_fd);

	}
    }

  close (fd);
  log_quit ();

  signal_on ();

  return 0;
}

static void
usage (void)
{
  fprintf (stdout, "%s %s - %s\n\n", PACKAGE, VERSION, AUTHOR);

  fprintf (stdout, "\t-f or --file <file>         Set your config file\n");
  fprintf (stdout, "\t-l or --logfile <file>      Set your log file\n");
  fprintf (stdout, "\t-b or --background <file>   0 no background, 1 yes\n");
  fprintf (stdout, "\t-d or --debug <int>         Set your debug level:\n"
	   "\t                            - 0 only error\n"
	   "\t                            - 1 error and warning\n"
	   "\t                            - 2 all messages\n");
  fprintf (stdout, "\t-i or --input <str>         Set your input element\n");
  fprintf (stdout, "\t-o or --output <str>        Set your output element\n");
  fprintf (stdout, "\nInput element syntax is:\n"
	   "\ttype=t/u/c/f or tcp/unix/command/filedescriptor\n"
	   "\tservername=0.0.0.0\n"
	   "\tport=12345\n"
	   "\tunixpath=/tmp/sock.erd\n"
	   "\tlisten=1000\n"
	   "\tssl=0/1\n"
	   "\tcertificate=certificate.pem\n"
	   "\tprivatekey=privatekey.pem\n"
	   "\tcommand=/bin/ls\n");
  fprintf (stdout, "\nOutput element syntax is:\n"
	   "\ttype=t/u/c/f or tcp/unix/command/filedescriptor\n"
	   "\tserver=www.google.it\n"
	   "\tport=80\n"
	   "\tunixpath=/tmp/sock.erd\n"
	   "\tssl=0/1\n"
	   "\tcommand=/bin/ls\n");

  fprintf (stdout, "\n");
}

/* EOF */
