/* ERD - Copyright (C) 2005 bakunin - Andrea Marchesini 
 *                              <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "erd.h"

static size_t erd_read (struct erd_item *, char *, size_t);
static size_t erd_write (struct erd_item *, char *, size_t);

void
erd_redirect (struct erd_config *conf, struct erd_item *c)
{
  int fd = 0;
  char buff[1024];
  struct timeval tv;
  fd_set in;
  int ret;
  struct erd_item *sock;
  pid_t pid = 0;
  int fd_input=0;

  SSL_CTX *ctx = NULL;
  SSL_METHOD *meth;
  SSL *ssl = NULL;

  /* Create socket... */
  if (conf->output_type == ERD_TCPSOCKET)
    {

      struct sockaddr_in s;
      struct hostent *hp;

      if ((fd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
	{
	  log_write (ERD_INFO, "Socket error");
	  return;
	}

      if (!(hp = gethostbyname (conf->output_server)))
	{
	  close (fd);
	  log_write (ERD_INFO, "Host %s unknowed", conf->output_server);
	  return;
	}

      memset ((void *) &s, 0, sizeof (s));
      s.sin_family = AF_INET;
      s.sin_addr.s_addr = ((struct in_addr *) (hp->h_addr))->s_addr;
      s.sin_port = htons (conf->output_port);

      if (connect (fd, (struct sockaddr *) &s, sizeof (s)) < 0)
	{
	  close (fd);
	  log_write (ERD_INFO, "Connect error to %s", conf->output_server);
	  return;
	}
    }

  /* Unix socket */
  else if (conf->output_type == ERD_UNIXSOCKET)
    {
      struct sockaddr_un saddr;

      if ((fd = socket (AF_UNIX, SOCK_STREAM, 0)) < 0)
	{
	  log_write (ERD_INFO, "Socket error");
	  return;
	}

      saddr.sun_family = AF_UNIX;
      strncpy (saddr.sun_path, conf->output_unixpath,
	       sizeof (saddr.sun_path));

      if (connect (fd, (struct sockaddr *) &saddr, sizeof (saddr)) < 0)
	{
	  close (fd);
	  log_write (ERD_INFO, "Connect error to %s", conf->output_unixpath);
	  return;
	}
    }

  /* Command */
  else if (conf->output_type == ERD_COMMAND)
    {
      char **arg = command_parse (conf->output_command);
      int p_in[2];
      int p_out[2];

      if (pipe (p_in))
	fatal ("Pipe error.");

      if (pipe (p_out))
	fatal ("Pipe error.");

      switch ((pid = fork ()))
	{
	case -1:
	  fatal ("Forking error.");
	  break;

	case 0:
	  close (p_out[0]);
	  close (STDOUT_FILENO);

	  close (p_in[1]);
	  close (STDIN_FILENO);

	  dup2 (p_in[0], STDIN_FILENO);
	  dup2 (p_out[1], STDOUT_FILENO);

	  execve (arg[0], arg, conf->env);

	  close (p_out[0]);
	  close (p_in[0]);
	  exit (1);

	  break;

	default:
	  close (p_out[1]);
	  close (p_in[0]);

	  fd = p_out[0];
	  fd_input=p_in[1];
	}

      command_free_arg (arg);
    }

  /* FD */
  else if (conf->output_type == ERD_FD)
    {
       fd = STDOUT_FILENO;
    }

  /* Active the connection for ssl */
  if (conf->output_ssl)
    {

      meth = SSLv23_client_method ();
      if (!(ctx = SSL_CTX_new (meth)))
	{
	  close (fd);
	  return;
	}

      if ((ssl = SSL_new (ctx)) < 0)
	{
	  close (fd);
	  return;
	}

      SSL_set_fd (ssl, fd);
      if (SSL_connect (ssl) < 0)
	{
	  close (fd);
	  SSL_shutdown (ssl);
	  SSL_free (ssl);
	  SSL_CTX_free (ctx);
	  return;
	}
    }

  if (!(sock = (struct erd_item *) malloc (sizeof (struct erd_item))))
    fatal ("Error: memory.");

  sock->fd = fd;
  sock->ssl = ssl;
  sock->use_ssl = conf->output_ssl;
  sock->pid = pid;

  while (1)
    {
      FD_ZERO (&in);
      FD_SET (c->fd, &in);
      FD_SET (fd, &in);

      tv.tv_sec = 0;
      tv.tv_usec = 10000;

      if (c->pid && waitpid (c->pid, NULL, WNOHANG) == c->pid)
	break;

      if (sock->pid && waitpid (sock->pid, NULL, WNOHANG) == sock->pid)
	break;

      if (c->fd > sock->fd)
	ret = select (c->fd + 1, &in, NULL, NULL, &tv);
      else
	ret = select (sock->fd + 1, &in, NULL, NULL, &tv);

      if (ret < 0 && errno != EINTR)
	{
	  erd_close (sock);
	  if (ctx)
	    SSL_CTX_free (ctx);
	  return;
	}

      if (FD_ISSET (c->fd, &in))
	{
	  ret = erd_read (c, buff, sizeof (buff));

	  if (ret < 0 && errno == EINTR)
	    {
	      erd_close (sock);
	      if (ctx)
		SSL_CTX_free (ctx);
	      return;
	    }
	  else if (ret == 0)
	    {
	      erd_close (sock);
	      if (ctx)
		SSL_CTX_free (ctx);
	      return;
	    }

	  if (!pid && erd_write (sock, buff, ret) != ret)
	    {
	      erd_close (sock);
	      if (ctx)
		SSL_CTX_free (ctx);
	      return;
	    }

	  if (pid && write (fd_input, buff, ret) != ret)
	    {
	      erd_close (sock);
	      if (ctx)
		SSL_CTX_free (ctx);
	      return;
	    }
	}

      else if (FD_ISSET (fd, &in))
	{
	  ret = erd_read (sock, buff, sizeof (buff));

	  if (ret < 0 && errno == EINTR)
	    {
	      erd_close (sock);
	      if (ctx)
		SSL_CTX_free (ctx);
	      return;
	    }
	  else if (ret == 0)
	    {
	      erd_close (sock);
	      if (ctx)
		SSL_CTX_free (ctx);
	      return;
	    }

	  if (!c->pid && erd_write (c, buff, ret) != ret)
	    {
	      erd_close (sock);
	      if (ctx)
		SSL_CTX_free (ctx);
	      return;
	    }

	  else if (c->pid && write (c->fd_input, buff, ret) != ret)
	    {
	      erd_close (sock);
	      if (ctx)
		SSL_CTX_free (ctx);
	      return;
	    }
	}
    }

  free (sock);
  if (ctx)
    SSL_CTX_free (ctx);
}

static size_t
erd_write (struct erd_item *c, char *buf, size_t size)
{
  if (c->use_ssl)
    return SSL_write (c->ssl, buf, size);

  return write (c->fd, buf, size);
}

static size_t
erd_read (struct erd_item *c, char *buf, size_t size)
{
  if (c->use_ssl)
    return SSL_read (c->ssl, buf, size);

  return read (c->fd, buf, size);
}

void
erd_close (struct erd_item *c)
{
  if (c->use_ssl)
    {
      SSL_shutdown (c->ssl);
      SSL_free (c->ssl);
    }

  shutdown (c->fd, SHUT_RDWR);
  close (c->fd);

  free (c);
}

/* EOF */
